import numpy as np
from einsum_pipe import einsum_pipe, compile_einsum_args
from einsum_pipe.einsum_script import EinsumScript


def einsum_pipe_simple(*args):
    subs = [arg for arg in args if isinstance(arg, (str, list, tuple))]
    ops = [arg for arg in args if not isinstance(arg, (str, list, tuple))]

    state: np.ndarray = ops.pop(0)
    for sub in subs:
        if isinstance(sub, str):
            extra_state = [ops.pop(0) for _ in range(sub.count(','))]
            state = np.einsum(sub, state, *extra_state)
        else:
            state = np.reshape(state, sub)

    return state


def test_single_arg_no_reshape():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc->bca',
        'abc->ab',
        'ab->a',
        A
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_single_arg_with_reshape():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc->bca',
        (10, 2, 5, 3, 2, 10),
        'abdfbg->afg',
        'abc->a',
        A
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_multi_args():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,dec->bcad',
        (10, 2, 5, 3, 2, 10, 5, 2),
        'abcdbace->eacd',
        'abcd->acd',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_mode():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,dec',
        'acbd',
        'cbdd',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_broadcasting():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,...c->bca...',
        (10, 2, 5, 3, 2, 10, 5, 2, 20),
        'abcdbace...->eacd...',
        'abcd...->acd...',
        'acb...',
        'c...ba',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_unequal_broadcasting():
    A = np.random.rand(10, 20, 20)
    B = np.random.rand(10, 20, 30)
    args = [
        'ab...,...c->bca...',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_array_creation():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc,->bca',
        'abc->a',
        A, 10
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_multistage_args():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'bac',
        'abc,dec->bcad',
        (10, 2, 5, 3, 2, 10, 5, 2),
        'abcdbace,->eacd',
        'abcd->acd',
        A, B, 10
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_broadcast():
    A = np.random.rand(10)
    B = np.random.rand(10)
    args = [
        '...,...',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_basic_simplify():
    script_a = EinsumScript.parse([[10, 20, 5, 6]], 'abcd->acd')
    script_a.simplify()
    script_b = EinsumScript.parse([[10, 20, 30]], 'abc->ac')
    assert str(script_a) == str(script_b)


def test_simplify():
    # Make a discontiguous array
    A = np.random.rand(3, 3, 3, 9).transpose((0, 2, 1, 3))
    assert not A.flags['CONTIGUOUS']

    (script, *_), output_shape = compile_einsum_args(
        [(27, 3, 3), 'abc->bc'], [A.shape], True)
    (script_max, *_), output_shape_max = compile_einsum_args(
        [(27, 3, 3), 'abc->bc'], [A.shape], 'max')

    X = A.reshape(script.input_shapes[0])
    Y = A.reshape(script_max.input_shapes[0])
    assert all(x == y for x, y in zip(output_shape, output_shape_max))
    assert np.shares_memory(A, X) and not np.shares_memory(A, Y)
    assert np.allclose(einsum_pipe(A, script=script, output_shape=output_shape), einsum_pipe(
        A, script=script_max, output_shape=output_shape_max))


def test_incompatible_steps_simple():
    A = np.random.rand(6)
    args = [
        [3, 2],
        'ij->ji',
        [3, 2],
        'ij->ji',
        A
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_incompatible_steps_many():
    A = np.random.rand(6, 4, 12)
    B = np.random.rand(4, 4, 12)
    args = [
        [3, 2, 4, 12],
        'ijkl->ji',
        [6],
        'i,klm->iklm',
        [3, 2, 4, 4, 12],
        'ijklm->jiklm',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))



from __future__ import annotations
import copy
import math
from typing import Generator, List, Optional, Tuple, TypeVar


class EinsumComp:
    def __init__(self, size: int, parsed_from: Optional[str] = None) -> None:
        self.size = size
        self._parsed_from = parsed_from


class IncompatibleShapeError(Exception):
    def __init__(self, shape_a: Tuple[int, ...], shape_b: Tuple[int, ...]) -> None:
        message = f'Incompatible shapes: {shape_a} vs {shape_b}'
        super().__init__(message)


def _get_char(index: int) -> str:
    assert index < 26*2
    return chr((ord('a') if index < 26 else (ord('A') - 26)) + index)


class EinsumScript:
    def __init__(self, inputs: List[List[EinsumComp]], outputs: List[EinsumComp]) -> None:
        self.inputs = inputs
        self.outputs = outputs
        self._parsed_script: Optional[str] = None

    @classmethod
    def parse(cls, input_shapes: List[List[int]], subscripts: str) -> EinsumScript:
        parsed_subscripts = subscripts
        subscripts = subscripts.replace(' ', '')
        # Easier to deal with broadcasting as a single character
        subscripts = subscripts.replace('...', '.')
        # The broadcasting character is automatically sorted to the start
        letters = sorted(subscripts.replace(',', '').replace('->', ''))
        if '->' not in subscripts:
            output_letters = [l for l in letters if l ==
                              '.' or letters.count(l) == 1]
            if (bc_count := output_letters.count('.')) > 1:
                output_letters = output_letters[bc_count - 1:]
            subscripts += '->' + ''.join(output_letters)
        letter_dict = {v: EinsumComp(0, v) for v in set(letters) if v != '.'}

        inputs_subs, output_subs = subscripts.split('->')
        inputs: List[List[EinsumComp]] = []
        broadcast_comps: List[EinsumComp] = []
        assert len(inputs_subs.split(',')) == len(
            input_shapes), f'''Error while parsing "{subscripts}" with input shapes {input_shapes}: insuficient input shapes found for number of references in subscripts!'''
        for sub, shape in zip(inputs_subs.split(','), input_shapes):
            inputs.append([])
            for c in sub:
                if c == '.':
                    # Broadcasting works from the last axis to the first and shares these axes with other broadcasts
                    undefined_axes = len(shape) - (len(sub) - 1)
                    for _ in range(undefined_axes - len(broadcast_comps)):
                        broadcast_comps.insert(0, EinsumComp(0, '...'))
                    if undefined_axes > 0:
                        inputs[-1].extend(broadcast_comps[-undefined_axes:])
                else:
                    inputs[-1].append(letter_dict[c])

        assert len(inputs) > 0

        outputs: List[EinsumComp] = []
        for c in output_subs:
            if c == '.':
                # All broadcasted axes are added in order
                outputs.extend(broadcast_comps)
            else:
                outputs.append(letter_dict[c])

        script = EinsumScript(inputs, outputs)
        script._parsed_script = parsed_subscripts
        for inp, shape in zip(inputs, input_shapes):
            assert len(inp) == len(
                shape), f'Error while parsing "{subscripts}" with input shapes {input_shapes}: {len(inp)} != {len(shape)}'
            for comp, dim in zip(inp, shape):
                comp.size = dim

        return script

    def split_comp(self, comp: EinsumComp, part_sizes: List[int]) -> None:
        repeats = [EinsumComp(size, comp._parsed_from)
                   for size in part_sizes[1:]]
        comp.size = part_sizes[0]
        for inp in [*self.inputs, self.outputs]:
            for i in range(len(inp)-1, -1, -1):
                if inp[i] == comp:
                    for rep in repeats[::-1]:
                        inp.insert(i+1, rep)

    def remove_ones(self):
        for inp in [*self.inputs, self.outputs]:
            for i in range(len(inp)-1, -1, -1):
                if inp[i].size == 1:
                    inp.pop(i)

    @property
    def input_shapes(self) -> List[Tuple[int]]:
        return [tuple(comp.size for comp in inp) for inp in self.inputs]

    @property
    def output_shape(self) -> Tuple[int]:
        return tuple(comp.size for comp in self.outputs)

    def simplify(self):
        # Get sequences (repeated or not) in which each element is unique to the sequence
        # Might be more efficient with some sort of linked list, but doesn't matter
        seqs: List[Optional[EinsumComp]] = []
        for comps in [*self.inputs, self.outputs]:
            for prev_comp, comp, next_comp in zip([None, *comps[:-1]], comps,  [*comps[1:], None]):
                if comp not in seqs:
                    seqs.append(comp)
                else:
                    seqs.append(None)
                    i = seqs.index(comp)
                    if i == len(seqs) - 1 or seqs[i + 1] != next_comp:
                        seqs.insert(i + 1, None)
                    if i == 0 or seqs[i - 1] != prev_comp:
                        seqs.insert(i, None)
            seqs.append(None)

        groups: List[List[EinsumComp]] = [[]]
        for comp in seqs:
            if comp is None:
                groups.append([])
            else:
                groups[-1].append(comp)

        group_pairs = [(group, EinsumComp(math.prod(comp.size for comp in group), ''.join(
            comp._parsed_from or '.' for comp in group))) for group in groups if len(group) > 1]

        # To check if the sizes before reshaping are the same as the sizes after
        sizes_before = [math.prod(shape) for shape in [
            *self.input_shapes, self.output_shape]]

        # Replace sequences of comps with their respective group comp
        for comps in [*self.inputs, self.outputs]:
            for group, new_comp in group_pairs:
                while group[0] in comps:
                    i = comps.index(group[0])
                    comps[i] = new_comp
                    for _ in range(len(group) - 1):
                        comps.pop(i + 1)

        sizes_after = [math.prod(shape) for shape in [
            *self.input_shapes, self.output_shape]]
        assert all(before == after for before,
                   after in zip(sizes_before, sizes_after)), 'This is a bug. Please submit a bug report!'

    def simplified(self) -> EinsumScript:
        val = copy.deepcopy(self)
        val.simplify()
        return val

    def match_splits(self, shapes: List[Optional[Tuple[int, ...]]]) -> None:
        for inp, input_shape in zip(self.inputs, shapes):
            if input_shape is None:
                continue
            input_shape_iter = iter(input_shape[::-1])
            inp_in_iter = rev_mut_iter(inp)

            try:
                input_shape_val = next(input_shape_iter)
                inp_in_val = next(inp_in_iter)

                while True:
                    if input_shape_val == inp_in_val.size:
                        input_shape_val = next(input_shape_iter)
                        inp_in_val = next(inp_in_iter)
                    elif input_shape_val > inp_in_val.size:
                        input_shape_val //= inp_in_val.size
                        inp_in_val = next(inp_in_iter)
                    else:
                        self.split_comp(inp_in_val, [
                            inp_in_val.size // input_shape_val, input_shape_val])
                        input_shape_val = next(input_shape_iter)
            except StopIteration:
                pass

    def __repr__(self) -> str:
        if self._parsed_script is None:
            return f'"{self}"'
        else:
            return f'"{self}" (parsed from "{self._parsed_script}")'

    def __str__(self) -> str:
        # This is equivalent to using a set except that it preserves order (at least since Python 3.7)
        # This isn't required but produces more natural string outputs
        comps: List[EinsumComp] = list(dict.fromkeys(
            comp for inp in self.inputs for comp in inp).keys())

        subs = []
        for inp in self.inputs:
            subs.append(''.join(_get_char(comps.index(comp)) for comp in inp))

        output_str = ''.join(_get_char(comps.index(comp))
                             for comp in self.outputs)

        return ','.join(subs) + '->' + output_str

    def __add__(self, rhs: EinsumScript) -> EinsumScript:
        lhs = copy.deepcopy(self)
        rhs = copy.deepcopy(rhs)
        lhs_out_iter = rev_mut_iter(lhs.outputs)
        rhs_in_iter = rev_mut_iter(rhs.inputs[0])

        try:
            lhs_out_val = next(lhs_out_iter)
            rhs_in_val = next(rhs_in_iter)

            while True:
                if lhs_out_val.size == rhs_in_val.size:
                    lhs_out_val = next(lhs_out_iter)
                    rhs_in_val = next(rhs_in_iter)
                elif lhs_out_val.size % rhs_in_val.size == 0:
                    lhs.split_comp(lhs_out_val, [
                        lhs_out_val.size // rhs_in_val.size, rhs_in_val.size])
                    rhs_in_val = next(rhs_in_iter)
                elif rhs_in_val.size % lhs_out_val.size == 0:
                    rhs.split_comp(rhs_in_val, [
                        rhs_in_val.size // lhs_out_val.size, lhs_out_val.size])
                    lhs_out_val = next(lhs_out_iter)
                else:
                    raise IncompatibleShapeError(
                        lhs.output_shape, rhs.input_shapes[0])
        except StopIteration:
            pass

        rhs.remove_ones()
        lhs.remove_ones()

        assert len(lhs.outputs) == len(
            rhs.inputs[0]), f'Incompatible shapes between {repr(lhs)} and {repr(rhs)}: {lhs.output_shape} and {rhs.input_shapes[0]}'
        assert all(x.size == y.size for x, y in zip(
            lhs.outputs, rhs.inputs[0]))

        for i, x in enumerate(rhs.inputs[0]):
            val = lhs.outputs[i]
            lhs.outputs[i] = x
            for inp in lhs.inputs:
                if val in inp:
                    for j, y in enumerate(inp):
                        if y == val:
                            inp[j] = x

        return EinsumScript(lhs.inputs + rhs.inputs[1:], rhs.outputs)


T = TypeVar('T')


def rev_mut_iter(data: List[T]) -> Generator[T, None, None]:
    for i in range(len(data)-1, -1, -1):
        yield data[i]



import copy
from functools import reduce
import itertools
import math
from typing import Callable, Dict, List, Optional, Tuple, Union, cast, overload

import numpy as np
from .einsum_script import EinsumScript, IncompatibleShapeError
import opt_einsum

Shape = Tuple[int, ...]
Subscript = Union[Shape, str, Callable[[List[Shape]],
                                       'Subscript'], List['Subscript'], Tuple['Subscript', ...]]


class _ScriptAdder:
    def __init__(self, scripts: List[EinsumScript]) -> None:
        self.scripts = scripts
        self.cache: Dict[Tuple[int, int], EinsumScript] = {}

    def __call__(self, start: int, stop: int) -> EinsumScript:
        try:
            (cache_start, cache_stop), output = max(((key, val) for key, val in self.cache.items(
            ) if key[0] >= start and key[1] <= stop), key=lambda item: item[0][1] - item[0][0])
            if cache_start != start:
                output = self(start, cache_start) + output
            if cache_stop != stop:
                output += self(cache_stop, stop)
            return output
        except ValueError:
            return reduce(lambda x, y: x+y, self.scripts[start:stop])


def _improve_splits(scripts: List[EinsumScript], output_sizes: List[int], split_indices: List[int], script_adder: _ScriptAdder) -> List[int]:
    # Find the combinations of indices which would result in the smallest total memory
    combinations = itertools.product(
        *[list(range(x+1, y+1)) for x, y in zip([0, *split_indices], split_indices)])
    sorted_groups = sorted(combinations, key=lambda comb: sum(
        output_sizes[x-1] for x in comb))

    for indices in sorted_groups:
        try:
            for x, y in zip([0, *indices], [*indices, len(scripts)]):
                script_adder(x, y)
            return list(indices)
        except IncompatibleShapeError:
            pass
    return split_indices


def _collapse(scripts: List[EinsumScript]) -> List[EinsumScript]:
    if len(scripts) == 0:
        return []
    split_indices: List[int] = []
    splits: List[EinsumScript] = [scripts[0]]
    output_sizes: List[int] = []
    for i, script in enumerate(scripts[1:], 1):
        script.simplify()
        try:
            output_sizes.append(math.prod(splits[-1].output_shape))
            splits[-1] += script
        except IncompatibleShapeError:
            splits.append(script)
            split_indices.append(i)

    split_min_sizes = [min(output_sizes[x:y])
                       for x, y in zip([None, *split_indices], split_indices)]
    if len(splits) == 1 or all(output_sizes[i-1] == split_min_size for i, split_min_size in zip(split_indices, split_min_sizes)):
        return splits
    else:
        script_adder = _ScriptAdder(scripts)
        split_indices = _improve_splits(
            scripts, output_sizes, split_indices, script_adder)
        return [script_adder(x, y) for x, y in zip([0, *split_indices], [*split_indices, len(scripts)])]


def compile_einsum_args(subscripts: List[Subscript], input_shapes: List[Tuple[int, ...]],
                        simplify: Union[str, bool] = True) -> Tuple[List[EinsumScript], Tuple[int, ...]]:
    unused_shapes = copy.copy(input_shapes)
    scripts: List[EinsumScript] = []

    while len(subscripts) > 0:
        sub = subscripts.pop(0)
        if isinstance(sub, str):
            # Normal subscript
            nargs = sub.count(',') + 1
            next_input_shapes = [list(unused_shapes.pop(0))
                                 for _ in range(nargs)]
            script = EinsumScript.parse(next_input_shapes, sub)
            unused_shapes.insert(0, script.output_shape)
            scripts.append(script)
        elif callable(sub):
            # Lazy argument
            subscripts.insert(0, sub(unused_shapes))
        elif isinstance(sub, (list, tuple)):
            if isinstance(sub[0], int):
                # Reshape
                unused_shapes[0] = cast(Tuple[int], tuple(sub))
            else:
                # Inner list which needs to be flattened
                for val in sub[::-1]:
                    subscripts.insert(0, cast(Subscript, val))

    output_shape = unused_shapes[0]
    output_scripts = _collapse(scripts)

    if simplify == 'max':
        for script in output_scripts:
            script.simplify()
    elif simplify:
        input_shape_iter = iter(input_shapes)
        first_val = next(input_shape_iter)
        for script in output_scripts:
            script.simplify()
            script.match_splits([first_val] + [next(input_shape_iter)
                                for _ in script.inputs[1:]])
            first_val = None
    return output_scripts, output_shape


@overload
def einsum_pipe(*args, simplify=True, **kwargs) -> np.ndarray: ...


@overload
def einsum_pipe(*args, simplify=True,
                script: EinsumScript, output_shape: Tuple[int, ...], **kwargs) -> np.ndarray: ...


def einsum_pipe(*args, simplify=True,
                script: Optional[EinsumScript] = None, output_shape: Optional[Tuple[int, ...]] = None, **kwargs) -> np.ndarray:
    assert (script is None and output_shape is None) or (
        script is not None and output_shape is not None)
    subs = []
    ops: List[np.ndarray] = []
    for arg in args:
        if isinstance(arg, (str, list, tuple)) or callable(arg):
            if isinstance(arg, list) and not isinstance(arg[0], int):
                subs.extend(arg)
            else:
                subs.append(arg)
        else:
            try:
                assert arg.shape is not None
                ops.append(arg)
            except AttributeError:
                ops.append(np.array(arg))

    if script is None:
        output_scripts, output_shape = compile_einsum_args(
            subs, [op.shape for op in ops], simplify=simplify)
    else:
        output_scripts = [script]

    ops_iter = iter(ops)
    state = next(ops_iter)

    for script in output_scripts:
        reshaped_ops = [np.reshape(op, shape)
                        for shape, op in zip(script.input_shapes, itertools.chain([state], ops_iter))]
        state = cast(np.ndarray, opt_einsum.contract(
            str(script), *reshaped_ops, **kwargs))

    return state.reshape(cast(Shape, output_shape))
